// (c) Bloodymad 07/2003
// Partially based on code by io^2
// Uses code from Mark Adler

#include "StdAfx.h"
#include "VirtualZipFile.h"
#include "otherfunctions.h"


// Implementation CVZSection

CVZSection::CVZSection() {
	sectionoffset = 0;
	sectionactive = false;
}

CVZSection::~CVZSection() {
}

bool CVZSection::OffsetInSection(uint32 offset) const {
	return (offset>=Begin()) && (offset<=End());
}

void CVZSection::EnterSection() {
	sectionactive = true;
	LocalSeek(0);
}

void CVZSection::LeaveSection() {
	sectionactive = false;
}

void CVZSection::Seek(uint32 lOff) {
	ASSERT(sectionactive);
	if (!OffsetInSection(lOff))
		throw new CFileException(CFileException::badSeek);
	LocalSeek(lOff-sectionoffset);
}


// CVZPositionedSection (abstract)

class CVZPositionedSection : public CVZSection { // abstract
protected:
	uint32 pos;
	uint32 size;
	uint32 MaxRead() const;
public:
	CVZPositionedSection();
	CVZPositionedSection(uint32 asize);
	void LocalSeek(uint32 lOff);
	uint32 Size() const;
};

// Implementation CVZPositionedSection

uint32 CVZPositionedSection::MaxRead() const {
	return Size()-pos;
}

CVZPositionedSection::CVZPositionedSection() {
	pos = 0;
	size = 0;
}

CVZPositionedSection::CVZPositionedSection(uint32 asize) {
	pos = 0;
	size = asize;
}

void CVZPositionedSection::LocalSeek(uint32 lOff) {
	pos = lOff;
}

uint32 CVZPositionedSection::Size() const {
	return size;
}

// CVZBufferSection

class CVZBufferSection : public CVZPositionedSection {
protected:
	byte	*data;
public:
	CVZBufferSection(uint32 secoffset, void *buf, uint32 bsize);
	~CVZBufferSection();
	uint32 Read(void* lpBuf, uint32 nCount);
};

// Implementation CVZBufferSection

CVZBufferSection::CVZBufferSection(uint32 secoffset, void *buf, uint32 bsize) {
	sectionoffset = secoffset;
	data = NULL;
	data = (byte*)malloc(bsize);
	memcpy(data,buf,bsize);
	size = bsize;
}

CVZBufferSection::~CVZBufferSection() {
	if (data) {
		delete data;
		data = NULL;
	}
}

uint32 CVZBufferSection::Read(void* lpBuf, uint32 nCount) {
	ASSERT(sectionactive);
	uint32 actread = MaxRead()>nCount ? nCount : MaxRead();
	memcpy(lpBuf,data+pos,actread);
	pos += actread;
	return actread;
}

// CVZZeroSection

class CVZZeroSection : public CVZPositionedSection {
public:
	CVZZeroSection(uint32 secoffset, uint32 secsize);
	uint32 Read(void* lpBuf, uint32 nCount);
};

// Implementation CVZZeroSection

CVZZeroSection::CVZZeroSection(uint32 secoffset, uint32 secsize) {
	sectionoffset = secoffset;
	size = secsize;
}

uint32 CVZZeroSection::Read(void* lpBuf, uint32 nCount) {
	ASSERT(sectionactive);
	uint32 actread = MaxRead()>nCount ? nCount : MaxRead();
	memset(lpBuf,0,actread);
	pos += actread;
	return actread;
}


// CVZFileSection

class CVZFileSection : public CVZSection {
protected:
	UINT openflags;
	char *filename;
	uint32 size;
	CFile *data;
	uint32 MaxRead() const;
public:
	CVZFileSection(uint32 secoffset, const char* afilename, UINT nOpenFlags);
	~CVZFileSection();
	void EnterSection();
	void LeaveSection();
	void LocalSeek(uint32 lOff);
	uint32 Read(void* lpBuf, uint32 nCount);
	uint32 Size() const;
};

CVZFileSection::CVZFileSection(uint32 secoffset, const char *afilename, UINT nOpenFlags) {
	ASSERT(afilename);
	data = NULL;
	filename = NULL;
	openflags = nOpenFlags;
	sectionoffset = secoffset;
	filename = nstrdup(afilename);
	CFileStatus status;
	if (!CFile::GetStatus(afilename,status))
		throw new CFileException(CFileException::fileNotFound);
	size = status.m_size;
}

CVZFileSection::~CVZFileSection() {
	if (filename) {
		delete filename;
		filename = NULL;
	}
	if (data) {
		delete data;
		data = NULL;
	}
}

uint32 CVZFileSection::MaxRead() const {
	if (!sectionactive)
		return size;
	ASSERT(data);
	return size-data->GetPosition();
}

void CVZFileSection::EnterSection() {
	if (sectionactive) {
		ASSERT(data);
		data->Seek(0,CFile::begin);
	}
	else {
		ASSERT(!data);
		data = new CFile(filename,openflags);
	}
	CVZSection::EnterSection();
}

void CVZFileSection::LeaveSection() {
	if (sectionactive) {
		ASSERT(data);
		delete data;
		data = NULL;
	}
	CVZSection::LeaveSection();
}

void CVZFileSection::LocalSeek(uint32 lOff) {
	data->Seek(lOff,CFile::begin);
}

uint32 CVZFileSection::Read(void* lpBuf, uint32 nCount) {
	ASSERT(sectionactive);
	ASSERT(data);
	return data->Read(lpBuf,nCount);
}

uint32 CVZFileSection::Size() const {
	return size;
}


//////
// Cyclic Redundancy Check - Copyright (C) 1995-1998 Mark Adler

uint32 crcTable[256] = {
	0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
	0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
	0xe0d5e91eL, 0x97d2d988L, 0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L,
	0x90bf1d91L, 0x1db71064L, 0x6ab020f2L, 0xf3b97148L, 0x84be41deL,
	0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L, 0x136c9856L,
	0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
	0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L,
	0xa2677172L, 0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL,
	0x35b5a8faL, 0x42b2986cL, 0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L,
	0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L, 0x26d930acL, 0x51de003aL,
	0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L, 0xcfba9599L,
	0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
	0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L,
	0x01db7106L, 0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL,
	0x9fbfe4a5L, 0xe8b8d433L, 0x7807c9a2L, 0x0f00f934L, 0x9609a88eL,
	0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL, 0x91646c97L, 0xe6635c01L,
	0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL, 0x6c0695edL,
	0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
	0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L,
	0xfbd44c65L, 0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L,
	0x4adfa541L, 0x3dd895d7L, 0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL,
	0x346ed9fcL, 0xad678846L, 0xda60b8d0L, 0x44042d73L, 0x33031de5L,
	0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL, 0xbe0b1010L,
	0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
	0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L,
	0x2eb40d81L, 0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L,
	0x03b6e20cL, 0x74b1d29aL, 0xead54739L, 0x9dd277afL, 0x04db2615L,
	0x73dc1683L, 0xe3630b12L, 0x94643b84L, 0x0d6d6a3eL, 0x7a6a5aa8L,
	0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L, 0xf00f9344L,
	0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
	0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL,
	0x67dd4accL, 0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L,
	0xd6d6a3e8L, 0xa1d1937eL, 0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L,
	0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL, 0xd80d2bdaL, 0xaf0a1b4cL,
	0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L, 0x316e8eefL,
	0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
	0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL,
	0xb2bd0b28L, 0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L,
	0x2cd99e8bL, 0x5bdeae1dL, 0x9b64c2b0L, 0xec63f226L, 0x756aa39cL,
	0x026d930aL, 0x9c0906a9L, 0xeb0e363fL, 0x72076785L, 0x05005713L,
	0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L, 0x92d28e9bL,
	0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
	0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L,
	0x18b74777L, 0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL,
	0x8f659effL, 0xf862ae69L, 0x616bffd3L, 0x166ccf45L, 0xa00ae278L,
	0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L, 0xa7672661L, 0xd06016f7L,
	0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL, 0x40df0b66L,
	0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
	0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L,
	0xcdd70693L, 0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L,
	0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
	0x2d02ef8dL
};
	
uint32 updateCRC( uint32 crc, const uint8 *buf, uint32 len ) {
	uint32 c = crc;
	
	for( uint32 n = 0; n < len; n++ )
		c = crcTable[( c ^ buf[n])&0xff ] ^ ( c>>8 );

	return( c );
}

inline uint32 calcCRC( const uint8 *buf, uint32 len ) {
	return( updateCRC( 0xffffffff, buf, len )^0xffffffff );
}

//////
// FileMapR

	class FileMapR {
	public:
		FileMapR( const char *fileName ) : hFile(0), hMap(0), mem(0), fileSize(0) {
			hFile     = ::CreateFile( fileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );
			ASSERT(hFile && hFile!=INVALID_HANDLE_VALUE);
			fileSize  = ::GetFileSize( hFile, 0 );
			hMap      = ::CreateFileMapping( hFile, NULL, PAGE_READONLY, 0,0, NULL );
			ASSERT(hMap && hMap!=INVALID_HANDLE_VALUE);
			mem       = ::MapViewOfFile( hMap, FILE_MAP_READ, 0,0,0 );
			ASSERT(mem);
		}

		~FileMapR() {
			if( mem )                                  ::UnmapViewOfFile( mem );
			if( hMap  && hMap!=INVALID_HANDLE_VALUE )  ::CloseHandle( hMap );
			if( hFile && hFile!=INVALID_HANDLE_VALUE ) ::CloseHandle( hFile );
		}

		inline uint8* data() {
			return( (uint8*)mem );
		}

		inline uint32 size() {
			return( fileSize );
		}


	private:
		HANDLE hFile;
		HANDLE hMap;
		void*  mem;
		uint32    fileSize;
	};


// Implementation CFileInfo

CFileInfo::CFileInfo() {
	zipcentraldir.signature              = 0x02014b50;
	zipcentraldir.version                = 0x14; // MSDOS, version 2.0
	zipcentraldir.minVersion             = 0x14; // ..
	zipcentraldir.flags                  = 0;
	zipcentraldir.compressionMethod      = 0;    // UNCOMPRESSED
	//zipcentraldir.lastFileMod            = info->mod;
	//zipcentraldir.crc32                  = info->crc;
	//zipcentraldir.compressedSize         = info->size;
	//zipcentraldir.uncompressedSize       = info->size;
	//zipcentraldir.filenameLength         = info->filenameLen;
	zipcentraldir.extrafieldLength       = 0;
	zipcentraldir.commentLength          = 0;
	zipcentraldir.disknumber             = 0;
	zipcentraldir.internalFileAttributes = 0;	   // binary
	//zipcentraldir.externalFileAttributes = info->attr;
	//zipcentraldir.localHeaderOffset      = offsets[block-num];
}

void CFileInfo::FillLocalFileHeader() {
	ziplocalheader.signature         = 0x04034b50;
	ziplocalheader.minVersion        = 0x14; // version 2.0
	ziplocalheader.flags             = 0;
	ziplocalheader.compressionMethod = 0;    // UNCOMPRESSED
	ziplocalheader.lastFileMod       = zipcentraldir.lastFileMod;
	ziplocalheader.crc32             = zipcentraldir.crc32;
	ziplocalheader.compressedSize    = zipcentraldir.compressedSize;
	ziplocalheader.uncompressedSize  = zipcentraldir.uncompressedSize;
	ziplocalheader.filenameLength    = zipcentraldir.filenameLength;
	ziplocalheader.extrafieldLength  = 0;
}

// CVirtualZipFile implementation

CVirtualZipFile::CVirtualZipFile() {
	sections = NULL;
	sections = new CList<CVZSection*,CVZSection*>();
	size = 0;
	position = 0;
	actsectionpos = 0;
	actsection = NULL;
	alignBlockSize = 1;
	basedir = NULL;
	openFlags= 0;
}

CVirtualZipFile::~CVirtualZipFile() {
	if (actsection)
		Close();
	while (!sections->IsEmpty())
		delete sections->RemoveTail();
	delete sections; sections = NULL;
	if (basedir) {
		delete basedir; basedir = NULL;
	}

}

CVirtualZipFile::CVirtualZipFile(LPCTSTR lpszDirName, UINT nOpenFlags, uint32 align) {
	ASSERT(AfxIsValidString(lpszDirName));
	if (!align)
		align = 1;
	sections = new CList<CVZSection*,CVZSection*>();
	alignBlockSize = align;
	actsectionpos = 0;
	actsection = NULL;

	CFileException e;
	if (!Open(lpszDirName, nOpenFlags, &e))
		AfxThrowFileException(e.m_cause, e.m_lOsError, e.m_strFileName);
}


void CVirtualZipFile::SwitchToSection(POSITION newsection) {
	actsection->LeaveSection();
	actsectionpos = newsection;
	actsection = sections->GetAt(newsection);
	actsection->EnterSection();
}


ULONGLONG CVirtualZipFile::GetPosition() const {
	return position;
}

CString CVirtualZipFile::GetFileName() const {
	return CString("VirtualZip");
}

BOOL CVirtualZipFile::Open(LPCTSTR lpszFileName, UINT nOpenFlags, CFileException* pException) {
	ASSERT_VALID(this);
	ASSERT(AfxIsValidString(lpszFileName));
	ASSERT(pException == NULL ||
		AfxIsValidAddress(pException, sizeof(CFileException)));

	// Shouldn't be already open
	ASSERT(!actsectionpos);

	// No write access:
	switch (nOpenFlags & 3) {
	case modeWrite:
	case modeReadWrite:
		throw new CFileException(CFileException::accessDenied);
	case modeRead:
		break;
	default:
		ASSERT(FALSE);  // invalid share mode
	}
	if (nOpenFlags & modeCreate)
		throw new CFileException(CFileException::accessDenied);

	openFlags = nOpenFlags;

	basedir = new CString(lpszFileName);
	CStringList searchdirs;
	CList<CFileInfo*,CFileInfo*> *fileinfos = new CList<CFileInfo*,CFileInfo*>();
	CFileFind filesearch;

	if (!basedir->IsEmpty() && basedir->GetAt(basedir->GetLength()-1)!='\\')
		basedir->Append("\\");

	searchdirs.AddHead("");
	while (!searchdirs.IsEmpty()) {
		CString subdir = searchdirs.RemoveHead();
		BOOL morefiles; 
		morefiles = filesearch.FindFile((*basedir)+subdir+"*.*");

		while (morefiles) {
			morefiles = filesearch.FindNextFile();
			if (!(filesearch.IsDots() || filesearch.IsHidden() || filesearch.IsTemporary())) {
				// We want this file / dir
				if (filesearch.IsDirectory()) {
					searchdirs.AddHead(subdir + filesearch.GetFileName() + "\\");
				}
				else if (filesearch.GetLength()<0x80000000) {
					// we can't handle bigger files
					CFileInfo* fileinfo = new CFileInfo();//basedir,subdir+filesearch.GetFileName());
					CString relname = subdir + filesearch.GetFileName();
					fileinfo->filename = nstrdup(relname);
					fileinfo->zipcentraldir.filenameLength = strlen(fileinfo->filename);
					CTime ctime;
					SYSTEMTIME systime;
					FILETIME filetime;
					uint32 dostime;
					filesearch.GetLastWriteTime(ctime); // May fail
					ctime.GetAsSystemTime(systime); // May fail
					::SystemTimeToFileTime(&systime,&filetime); // May fail
					::FileTimeToDosDateTime(&filetime, (uint16*)&dostime, ((uint16*)&dostime)+1); // May fail
					fileinfo->zipcentraldir.lastFileMod = dostime;
					fileinfo->zipcentraldir.uncompressedSize = filesearch.GetLength();
					fileinfo->zipcentraldir.compressedSize = filesearch.GetLength();
					fileinfo->zipcentraldir.externalFileAttributes =
						(filesearch.IsReadOnly() ? 1 : 0) ||
						(filesearch.IsHidden() ? 2 : 0) ||
						(filesearch.IsSystem() ? 4 : 0);

					FileMapR map((*basedir)+subdir+filesearch.GetFileName());
					fileinfo->zipcentraldir.crc32 = ::calcCRC(map.data(),map.size());

					fileinfo->FillLocalFileHeader();
					fileinfos->AddTail(fileinfo);
				}
			}
		} // while filesearch
		filesearch.Close();
	} // while searchdirs

	//actualinfo = fileinfos->GetHeadPosition();

	if (!fileinfos->IsEmpty()) {
		// gather tables
		uint64 zippos = 0;

		//int num = fileinfos->GetSize();
		
		//std::sort( fileList.begin(), fileList.end(), FileInfoSort());
		CVZSection *actsec;
		
		for (POSITION pos = fileinfos->GetHeadPosition();pos;fileinfos->GetNext(pos)) {
			CFileInfo *actinfo = fileinfos->GetAt(pos);
			actinfo->zipcentraldir.localHeaderOffset = (uint32)zippos;

			// Zip local file header
			actsec = new CVZBufferSection(zippos,&actinfo->ziplocalheader,sizeof(VZLocalFileHeader));
			sections->AddTail(actsec);
			zippos += actsec->Size();

			// Zip local file header file name
			actsec = new CVZBufferSection(zippos,actinfo->filename,strlen(actinfo->filename));
			sections->AddTail(actsec);
			zippos += actsec->Size();

			// Zip local file data
			actsec = new CVZFileSection(zippos,(*basedir)+actinfo->filename,nOpenFlags);
			sections->AddTail(actsec);
			zippos += actsec->Size();

			// Zip local file padding
			uint32 pad = ((zippos+alignBlockSize-1)/alignBlockSize)*alignBlockSize-zippos;
			if (pad) {
				actsec = new CVZZeroSection(zippos,pad);
				sections->AddTail(actsec);
				zippos += actsec->Size();
			}
		}
		uint32 cdirstart = zippos;

		// Central Directory
		for (POSITION pos = fileinfos->GetHeadPosition();pos;fileinfos->GetNext(pos)) {
			CFileInfo* actinfo = fileinfos->GetAt(pos);

			actsec = new CVZBufferSection(zippos,&actinfo->zipcentraldir,sizeof(VZCentralDirectory));
			sections->AddTail(actsec);
			zippos += actsec->Size();

			actsec = new CVZBufferSection(zippos,actinfo->filename,strlen(actinfo->filename));
			sections->AddTail(actsec);
			zippos += actsec->Size();
		}
		uint32 cdirend = zippos;
		
		VZCentralDirectoryEnd dirend;
		dirend.signature     = 0x06054b50;
		dirend.diskNo        = 0;
		dirend.diskCD        = 0;
		dirend.diskEntries   = fileinfos->GetCount();
		dirend.globalEntries = fileinfos->GetCount();
		dirend.cdLength      = cdirend-cdirstart; // length of all central dirs
		dirend.cdOffset      = cdirstart;                // first central dir
		dirend.commentLength = 0;

		actsec = new CVZBufferSection(zippos,&dirend,sizeof(dirend));
		sections->AddTail(actsec);
		zippos += actsec->Size();
		size = (uint32)zippos;

		if (zippos > 0xffffffff) {
			ASSERT(zippos <= 0xffffffff);//, "virtual zip exceeds 4GB limit" );
			return false;
		}
	}
	else { // sections->Empty()
		sections->AddTail(new CVZZeroSection(0,0));
		size = 0;
	}

	position = 0;
	actsectionpos = sections->GetHeadPosition();
	actsection = sections->GetAt(actsectionpos);
	actsection->EnterSection();

	return true;
}

CFile* CVirtualZipFile::Duplicate() const {
	// We just create a new one with the same parameters
	CVirtualZipFile *dup = new CVirtualZipFile(*basedir,openFlags,alignBlockSize);
	dup->Seek(position,begin);
	return dup;
};

ULONGLONG CVirtualZipFile::Seek(LONGLONG lOff, UINT nFrom) {
	ASSERT_VALID(this);
	ASSERT(m_hFile != INVALID_HANDLE_VALUE);
	ASSERT(nFrom == begin || nFrom == end || nFrom == current);

	if (nFrom==end) {
		if (lOff>size)
			throw new CFileException(CFileException::badSeek);
		lOff = size-lOff;
	}
	else if (nFrom==current) {
		lOff = position+lOff;
	}
	if (lOff>size)
		throw new CFileException(CFileException::badSeek);
	if (actsection->OffsetInSection(lOff))
		actsection->Seek(lOff);
	else if (lOff>position) {
		POSITION secp = actsectionpos;
		CVZSection *sec;
		do {
			sec = sections->GetNext(secp);
		} while (!sec->OffsetInSection(lOff));
		SwitchToSection(secp);
		actsection->Seek(lOff);
	}
	else {
		POSITION secp = actsectionpos;
		CVZSection *sec;
		do {
			sec = sections->GetNext(secp);
		} while (!sec->OffsetInSection(lOff));
		SwitchToSection(secp);
		actsection->Seek(lOff);
	}
	position = lOff;
	return position;
}

void CVirtualZipFile::SetLength(ULONGLONG dwNewLen) {
	if (dwNewLen!=size)
		// We don't change the length
		throw new CFileException(CFileException::accessDenied);
}

ULONGLONG CVirtualZipFile::GetLength() const {
	ASSERT_VALID(this);
	return size;
}

UINT CVirtualZipFile::Read(void* lpBuf, UINT nCount) {
	ASSERT_VALID(this);

	ASSERT(!nCount || lpBuf != NULL);
	ASSERT(!nCount || AfxIsValidAddress(lpBuf, nCount));

	UINT ncount = nCount;
	byte *lpbuf = (byte*)lpBuf;
	while (ncount) {
		// Read from the current section
		uint32 read = actsection->Read(lpbuf,ncount);
		ncount -= read;
		if (!ncount)
			// We have read everything we wanted
			break;

		// else: Get the next section
		POSITION nextsec = actsectionpos;
		sections->GetNext(nextsec);
		if (!nextsec)
			// There is no next section -> EoF
			break;

		// Prepare to read from next section
		lpbuf += read;
		SwitchToSection(nextsec);
	}
	return nCount-ncount;
}

void CVirtualZipFile::Write(const void* lpBuf, UINT nCount) {
	// We don't write
	throw new CFileException(CFileException::accessDenied);
}

void CVirtualZipFile::LockRange(ULONGLONG dwPos, ULONGLONG dwCount) {
	// We just don't so it.
}

void CVirtualZipFile::UnlockRange(ULONGLONG dwPos, ULONGLONG dwCount) {
	// We just don't do it.
}

void CVirtualZipFile::Abort() {
	ASSERT_VALID(this);
	if (actsection) {
		try {
			actsection->LeaveSection();
		}
		catch (...) {};
	}
}

void CVirtualZipFile::Flush() {
	// Because we don't write, we don't have to flush.
}

void CVirtualZipFile::Close() {
	if (!actsection)
		throw new CFileException(CFileException::invalidFile);
	actsection->LeaveSection();
	actsectionpos = 0;
	actsection = NULL;
	size = 0;
}
